# FILE: graphwin.tcl
#
# Entry graph canvas sub-widget
#
# The usual stacking order is:
#        Key (if any)
#        Horizontal and vertical rules
#        Axes and labels
#        Curves (if any)
#
# Two sets of coordinates are used: display (canvas, pixel) coordinates,
# and graph (data) coordinates.  Relevant variables for display coords
# are lmargin, tmargin, rmargin, bmargin, plotwidth and plotheight. Here
# lmargin+plotwidth+rmargin = [$graph cget -width], and similarly for
# canvas height.  The graph limits in data coords are held in xgraphmin,
# ygraphmin, xgraphmax, ygraphmax. The conversion between the two coords
# (x,y)->(cx,cy) is given by (cx,cy) = (xoff,yoff)+ (x*xscale,y*yscale).
#
# Oc_Class GraphWin methods:
#
#    callback method ResizeCanvas { width height }
#        Bound to the $graph <Configure> event.  Includes an
#      'update idletask' call.
#    method SetSmoothState { newstate }
#        If smooth state is 1, then use the "-smooth" option to 'canvas
#        create line' when plotting curves.  In this case the curves
#        will not generally pass through the data points.
#    method SetKeyState { newstate }
#    private method HideKey {}
#    method GetKeyPosition {}
#        Returns coords of upper righthand corner of current key, relative
#        to upper righthand corner of canvas.  (In particular, xpos will
#        generally be negative.)  The return value can be later fed back
#        into SetKeyState.
#    private method DrawKey { { coords {} } }
#        Routines for changing key (legend) display.
#    callback method ShowPosition { cx cy {convert 1}}
#    callback method UnshowPosition {}
#        Routines for displaying mouse positioning information.
#    private method DrawRules { x y }
#        Draws vertical rule at x=$x, and a horizontal rule at y=$y
#    callback method DragInit { tag x y }
#    callback method Drag { tag x y }
#        Bindings to allow the dragging of the key and rules.
#    method DrawAxes {}
#       Clears canvas, then renders title, axes labels, axes,
#       horizontal and vertical rules, and key.  Also sets
#       {t,b,l,r}margin, plotwidth/plotheight, and {x,y}scale,
#       {x,y}off values.
#    method NewCurve { curvename }
#    method DeleteCurve { curvename }
#    method DeleteAllCurves {}
#    private method GetCurveId { curvename } { id }
#       Curve creation, deletion and identification routines.
#    method AddDataPoints { curvename newpts }
#       Adds new points to data list, and marks display list
#       as out-of-date.  "newpts" is a list of the form
#       "x1 y1 x2 y2 x3 y3 ...".  Set x_i == {} to request
#       a break in the curve between points i-1 and i+1.
#       The return value is a bounding box for new data,
#       in graph coords, unless all x_i in newpts are {}, in
#       which case the return value is {}.
#    method DeleteAllData {}
#       Keeps curve structures intact, but clears all data,
#       and removes curves from display.
#    private method AddDisplayPoints { curveid datapts }
#       Called by UpdateDisplayList to append "datapts" to current
#       curveid curve piece.
#    private method UpdateDisplayList { curveid }
#       Makes the curve display list up-to-date with respect to
#       the data list.  The return value is the index into the
#       display list necessary to update the graphed display, i.e.,
#       the 'startindex' value appropriate for calling PlotCurve.
#       A value of -1 indicates that the curve should be completely
#       redrawn, and -2 occurs if no points were added to the display
#       list.
#    private method PlotCurve { curveid { startindex 0 }}
#       Draws specified portion (startindex to 'end') of specified
#       curve onto the canvas.
#    method GetCurveList {}
#       Returns an ordered list of all non-hidden curves (by name)
#    method DrawCurves {{update_only 1}}
#       Invokes PlotCurve for all non-hidden curves.  If update_only
#       is set, then only those portions of the curves not yet moved
#       to the display list are drawn.  After calling this function,
#       all non-hidden curve 'dispcurrent' elements are true.
#    private method InvalidateDisplay { curvelist }
#       Marks entire display portion of each curve in curvelist as
#       invalid.  Pass in $idlist to invalidate all curve displays.
#    private method GetDataExtents {}
#       Returns the minimal bounding box for all non-hidden curves.
#    private method RoundExtents {xmin ymin xmax ymax}
#       Expands values to "round" coordinates.
#    method SetGraphLimits {{xmin {}} {ymin {}} {xmax {}} {ymax {}}}
#       Allows client to set graph extents.  Empty-string valued imports
#       are filled with data from GetDataExtents, rounded outward using
#       RoundExtents.  The return value is 1 if the graph extents are
#       changed, 0 if the new values are the same as the old values.
#    method GetGraphLimits {}
#       Returns [list $xgraphmin $ygraphmin $xgraphmax $ygraphmin]
#    method RefreshDisplay {}
#       Redraws entire display, using current canvas size and graph
#       limits.
#    method Reset {}
#       Re-initializes most of the graph properties to initial defaults
#    method PrintData { curvepat }
#        For debugging.
#    method HasGui {}
#        Network interface.
#    method WinDestroy { w } { winpath }
#        $winpath <Destroy> binding.
#
# Non-class procs (used to cut down on call overhead) :
#   proc _IsInGraph { x y xmin ymin xmax ymax }
#     Returns a two element list, [list $xstat $ystat].  $xstat
#     is 0 if $xmin<=$x<=$xmax, -1 if $x<$xmin, and 1 if $x>max.
#     Analogous values for $ystat.
#   proc _PutInBox { x y xmin ymin xmax ymax } {
#     Returns a two element list [list $xnew $ynew], which shifts
#     x and y as necessary to put it into the specified box.

Oc_Class Ow_GraphWin {
    const public variable winpath
    public variable curve_width = 2  ;# Width of curve plot lines
    const public variable outer_frame_options = \
	    "-bd 0 -relief ridge -background #0000cd"
    # Note: #0000cd is MediumBlue
    const public variable default_height = 320
    const public variable default_width = 320
    const public variable canvas_frame_options = \
	    "-bd 0 -relief ridge -background #0000cd"
    const public variable canvas_options = \
	    "-bd 0 -relief flat -background white"
    const public variable graphsize = 0.9  ;# % of canvas to draw graph in.
    const public variable axescolor = black
    const public variable axeswidth = 2
    const public variable keywidth = 2

    const private variable segmax = 31 ;# Maximum number of curve updates
    ## to allow without a complete curve redraw.

    const private variable graph  ;# Graph canvas
    public method GetCanvas {} { return $graph }

    public variable title  = ""
    public variable title_font
    public variable xlabel = ""
    public variable ylabel = "" {
        # Separate ylabel characters by newlines
        set ylabel [join [split $ylabel {}] "\n"]
    }

    public variable label_font

    private variable hrulestate = 1      ;# Horizontal rule (1=>display),
    private variable hrulepos = {}       ;# position (in graph coords).
    private variable vrulestate = 1      ;# Vertical rule   (0=>hide)
    private variable vrulepos = {}       ;# position (in graph coords)
    ## h(v)rulepos == {} => draw (hide) rule at bottommost (leftmost)
    ## edge of the graph.

    public variable showkey = 1
    public variable smoothcurves = 0     ;# Draw curves as splines?
    private variable firstkey = 1        ;# If true, then a non-empty key
    ## has not been drawn in the lifetime of this widget.
    private variable key_font

    private variable lmargin = 0    ;# Margins: left, top, right, bottom,
    private variable tmargin = 0    ;# in pixels.
    private variable rmargin = 0
    private variable bmargin = 0
    private variable plotwidth  = 0 ;# Width and height of plotting region,
    private variable plotheight = 0 ;# in pixels.

    private variable curvecount = 0
    private variable xoff    = 0    ;# Used in conversion from data coords
    private variable yoff    = 0    ;#  to graph (display) coords.  Setup
    private variable xscale  = 1    ;#  in DrawAxes routine.
    private variable yscale  = -1
    private variable xgraphmin = 0  ;# Graph limits, in data coords.
    private variable xgraphmax = 1
    private variable ygraphmin = 0
    private variable ygraphmax = 1

    private variable but1x = 0       ;# Mouse coordinates when
    private variable but1y = 0       ;# Button-1 is pressed

    private array variable color     ;# Color array

    # Limits on number of points to store.  The maximum number of points
    # (per curve) is ptlimit.  When that value is exceeded, the list is
    # reduced to $ptlimit * $ptlimitcut.  Set ptlimit to 0 for no limit.
    public variable ptlimit = 0 {
        set ptlimit [expr {int(floor($ptlimit))}]
        if {$ptlimit<1} {
            set ptlimit 0
        } else {
            foreach curveid $idlist {
                $this CutCurve $curveid
            }
        }
    }
    private variable ptlimitcut = 0.8

    private variable idlist = {}     ;# Ordered list of all valid $curveid's
    private array variable id        ;# Name -> id mapping array
    private array variable curve     ;# Curve data array
    ## Each curve has an entry of the following form:
    ##       curve($curveid,name)     = id -> Name mapping
    ##       curve($curveid,color)    = Drawing color
    ##       curve($curveid,ptsbreak) = List of indices in pts of last
    ##             point in each "broken" segment of the display curve.
    ##             These result from x == {} entries in calls to
    ##             $this AddDataPoints.
    ##       curve($curveid,pts)      = Data point string of "x y" pairs
    ##       curve($curveid,lastpt)   = List pertaining to last data point
    ##             input:
    ##                { lastindex xlast ylast }
    ##             Here lastindex = expr [llength $curve($curveid,pts)]-2,
    ##             and { xlast ylast } =
    ##                        lrange $curve($curveid,pts) $lastindex end
    ##             This information is stored as an aid to Tcl 7.6, which
    ##             does not handle long lists very efficiently.
    ##       curve($curveid,xmin)     = Minimum value of x among data points
    ##       curve($curveid,xmax)     = Maximum value of x among data points
    ##       curve($curveid,ymin)     = Minimum value of y among data points
    ##       curve($curveid,ymax)     = Maximum value of y among data points
    ##       curve($curveid,ptsdisp)  = List of lists of display points.
    ##             Each separate "piece" of the curve is stored in its own
    ##             sublist.  Each sublist has at least 2 pairs of points.
    ##             These are all integer values.
    ##       curve($curveid,dnextpt)  = Index into curve($curveid,pts) of
    ##             next points to be processed for display.  Set to 0 if
    ##             pts has not been processed for display.
    ##       curve($curveid,dpiececount) = Number of "logical" curve sections
    ##             in curve($curveid,ptsdisp).  This will be more than the
    ##             number of elements of ptsdisp if any of the sections
    ##             contain fewer than 2 points.  (Sections with less than
    ##             2 points are dropped from ptsdisp so that all sublists in
    ##             ptsdisp can be directly fed to the canvas create line
    ##             command.)
    ##       curve($curveid,lastdpt)  = List pertaining to last display
    ##             processing pass:
    ##               { xprev yprev xstat ystat previnc }
    ##             'xprev yprev xstat ystat' are the data
    ##             position and the _IsInGraph return value of the last
    ##             "representative" point.  (If the trailing points in 'pts'
    ##             plot on the canvas at a distance less than 1 pixel, then
    ##             one "representative" point will be chosen for each display
    ##             pixel.  Because of this, 'xprev yprev' may occur earlier
    ##             in 'pts' than 'dnextpt-2'.)  'previnc' is true iff
    ##             T(xprev,yprev) is included in 'ptsdisp', where T represents
    ##             the transform from graph to display coordinates.  (All
    ##             positions in the 'lastpt' list are in graph coordinates.)
    ##             'previnc' is needed to handle the case where 'xprev yprev'
    ##             was excluded from the 'ptsdisp' because no part of the line
    ##             segment joining 'xprev yprev' to its predecessor lay inside
    ##             the graph viewport, but a portion of the line segment
    ##             between 'xprev yprev' and its successor does.  This can't
    ##             be determined until the successor is known, and so at that
    ##             time it may be necessary to retroactively add T(xprev,yprev)
    ##             to 'ptsdisp'.  lastdpt will be set to {} if a break in the
    ##             curve is to be positioned after the last entry in ptsdisp.
    ##       curve($curveid,dispcurrent)  = True iff ptsdisp is up-to-date
    ##                                  wrt pts and scaling.
    ##       curve($curveid,segcount) = Number of times PlotCurve has been
    ##             called for $curveid to perform only a partial display of
    ##             the curve.  This is typically equal to the number of
    ##             separate canvas items being used to display the curve.
    ##             When this number gets larger than segmax, PlotCurve
    ##             does a complete redraw rather than an update.
    ##       curve($curveid,hidden)   = Controls whether or not curve is drawn
    ##       curve($curveid,keyid)    = Canvas id associated with curve

    public variable callback = Oc_Nop
    public variable delete_callback_arg = {}

    Constructor {basewinpath args} {
	# Don't add '.' suffix if there already is one.
	# (This happens chiefly when basewinpath is exactly '.')
	if {![string match {.} \
		[string index $basewinpath \
		[expr {[string length $basewinpath]-1}]]]} {
	    append basewinpath {.}
	}

	set title_font [Oc_Font Get large_bold]
        set label_font [Oc_Font Get bold]
        set key_font $label_font

	eval $this Configure -winpath ${basewinpath}w$this $args

	# Temporarily disable callbacks until construction is complete
	set callback_keep $callback ;	set callback Oc_Nop

	# Pack entire subwidget into a subframe.  This makes
	# housekeeping easier, since the only bindings on
	# $winpath will be ones added from inside the class,
	# and also all subwindows can be destroyed by simply
	# destroying $winpath.
	eval frame $winpath $outer_frame_options
	set graph_frame [eval frame $winpath.canvasframe \
		$canvas_frame_options]
	set graph [eval canvas $graph_frame.graph \
		-width $default_width -height $default_height \
		$canvas_options]
	pack $graph -side left -anchor w -fill both -expand 1
	pack $graph_frame -side left -anchor w -fill both -expand 1 
	
	# Initialize color array
	array set color [list \
		0 #f00  \
		1 #0f0  \
		2 #00f  \
		3 #800  \
		4 #a0a  \
		5 #0aa  \
		6 #000  \
		7 #660  \
		8 #008  \
		9 #080   ]
	set color(count) 10
 
	# Widget bindings
	bind $winpath <Destroy> "$this WinDestroy %W"
	bind $graph <Configure> "$this ResizeCanvas %w %h"

	# Mouse bindings
	$graph bind showpos <Any-ButtonPress> "$this ShowPosition %x %y"
	$graph bind showpos <Any-ButtonRelease> "$this UnshowPosition"
	$graph bind key <ButtonPress-1>    "$this DragInit key %x %y"
	$graph bind hrule <ButtonPress-1>  "$this DragInit hrule %x %y"
	$graph bind vrule <ButtonPress-1>  "$this DragInit vrule %x %y"
	$graph bind key <Button1-Motion> \
                [Oc_SkipWrap "$this Drag key %x %y"]
	$graph bind hrule <Button1-Motion> \
                [Oc_SkipWrap "$this Drag hrule %x %y"]
	$graph bind vrule <Button1-Motion> \
                [Oc_SkipWrap "$this Drag vrule %x %y"]

	# Initialize scaling and axes
        $this SetGraphLimits
	$this RefreshDisplay

	# Re-enable callbacks
	set callback $callback_keep
    }
    callback method ResizeCanvas { width height } { graph } {
	set border [expr {2*[$graph cget -borderwidth] \
		+2*[$graph cget -highlightthickness]}]
	# Note: The <Configure> resize request *includes* both
	#       -borderwidth and -highlightthickness!

	set newwidth [expr {round($width-$border)}]
	set newheight [expr {round($height-$border)}]
	if {[$graph cget -width]!=$newwidth || \
		[$graph cget -height]!=$newheight} {
	    set confbind [bind $graph <Configure>]
	    bind $graph <Configure> {}  ;# Temporarily disable
	    $graph configure -width $newwidth -height $newheight
	    update idletasks  ;# Need this to keep from generating
	    ## and infinite <Configure> loop
	    bind $graph <Configure> $confbind ;# Reenable
	    $this RefreshDisplay
	}
    }
    method SetSmoothState { newstate } {
	if {$smoothcurves != $newstate} {
	    set smoothcurves $newstate
	    $this DrawCurves 0
	}
    }
    method SetKeyState { newstate {key_pos {}} } {
	set showkey $newstate
	if {$showkey} {
	    $this DrawKey $key_pos
	} else {
	    $this HideKey
	}
    }
    private method HideKey {} {
	$graph delete key
    }
    method GetKeyPosition {} {
        # Returns location of upper righthand corner of key, relative
        # to the upper righthand corner of the canvas.  The return
        # value can be fed into SetKeyState or DrawKey.
        set coords [$graph coords key_box]
	if {[string match {} $coords]} { return {} }
        set xpos [expr {[lindex $coords 2] - $lmargin \
                - $plotwidth - $rmargin}]
        set ypos [lindex $coords 1]
	return [list $xpos $ypos]
    }
    private method DrawKey { { key_pos {} } } {
	# Get coords from current key, if any, or otherwise
	# use graph extents
	if [string match {} $key_pos] {
	    set key_pos [$this GetKeyPosition]
	}

	$this HideKey   ;# Delete any previously existing key
	if { !$showkey } { return }

	set newbox 0 ;# If ==0, then we are overwriting old box
	if {$firstkey || [string match {} $key_pos]} {
	    set key_box_xmax [expr {$lmargin+$plotwidth}]
	    set key_box_ymin $tmargin
	    set newbox 1
	} else {
	    set key_box_xmax [expr {$lmargin + $plotwidth + \
                    $rmargin + [lindex $key_pos 0]}]
	    set key_box_ymin [lindex $key_pos 1]
	}

	# Temporarily write all key text items, in order, anchored
	# on ne corner; this is used to determine box dimensions.
	set keycount 0
        set linecount 0
	set x $key_box_xmax  ; set y $key_box_ymin
	foreach curveid $idlist {
	    if {! $curve($curveid,hidden) } {
		# Curve exists and is not hidden
		set tempid [$graph create text $x $y \
			-anchor ne \
			-fill $curve($curveid,color) \
			-justify left \
			-tags {key key_text} \
                        -font $label_font \
			-text $curve($curveid,name)]
		incr keycount
                incr linecount [expr {1 + \
                        [regsub -all "\n" $curve($curveid,name) {} dummy]}]
		set bbox [$graph bbox $tempid]
		set y [lindex $bbox 3]
	    }
	}

	# Create appropriately sized key box
	if {$keycount<1} {
	    set x1 $x  ; set y1 $y
	    set x2 $x1 ; set y2 $y1
	    set textheight 0
	    set textwidth 0
	    set lineheight 0
	} else {
	    set bbox [$graph bbox key_text]
	    set x1 [lindex $bbox 0] ; set y1 [lindex $bbox 1]
	    set x2 [lindex $bbox 2] ; set y2 [lindex $bbox 3]
	    set textheight [expr {$y2-$y1}]
	    set textwidth [expr {$x2-$x1}]
	    set lineheight [expr {double($textheight)/double($linecount)}]
	}
	set x $key_box_xmax  ; set y $key_box_ymin
	if {$newbox} {
	    # Add margin
	    set x [expr {$x-$lineheight}]
	    set y [expr {$y+0.75*$lineheight}]
	}
	set x1 [expr {$x-$textwidth-$lineheight/2.}]
	set y1 $y
	set x2 $x
	set y2 [expr {$y+$textheight+$lineheight/2.}]
        if {$x1<0} { set x2 [expr {$x2-$x1}] ; set x1 0 } ;# Don't
        ## allow key to be placed off lefthand side of graph.  This
        ## provides better behaved canvas resize behavior.
	if {$keycount<1} {
	    # Empty keybox
	    $graph create rectangle $x1 $y1 $x2 $y2 -tag {key key_box}\
		    -outline {}
	} else {
	    # Non-empty keybox
	    $graph create rectangle $x1 $y1 $x2 $y2 -tag {key key_box}\
		    -fill white -outline black -width $keywidth
	    set firstkey 0
	}

	# Rewrite all key text items, anchored on nw inside key box
	$graph delete key_text
	set x [expr {$x1+$lineheight/4.}]
	set y [expr {$y1+$lineheight/4.}]
	foreach curveid $idlist {
	    if {! $curve($curveid,hidden) } {
		# Curve exists and is not hidden
		set tempid [$graph create text $x $y \
			-anchor nw \
			-fill $curve($curveid,color) \
			-justify left \
			-tags {key key_text} \
                        -font $key_font \
			-text $curve($curveid,name)]
		set curve($curveid,keyid) $tempid
		set bbox [$graph bbox $tempid]
		set y [lindex $bbox 3]
	    }
	}
    }
    callback method ShowPosition { cx cy {convert 1}} \
            {graph xscale yscale xoff yoff xgraphmin xgraphmax \
             ygraphmin ygraphmax } {
	if {$convert} {
	    # Convert to canvas coordinates
	    set cx [$graph canvasx $cx]
	    set cy [$graph canvasy $cy]
	}

	# Convert to user coordinates
	if {$xscale==0. || $yscale==0.} {
	    set ux $cx  ;  set uy $cy   ;# Safety
	} else {
	    set ux [expr {($cx-$xoff)/double($xscale)}]
	    set uy [expr {($cy-$yoff)/double($yscale)}]
	}
	# Determine whether to display text to left or right
	# of point.
	set xmid [expr {($xgraphmin+$xgraphmax)/2.}]
	if {$ux>=$xmid} {
	    set textanchor se
	    set cxoff -3
	} else {
	    set textanchor sw
	    set cxoff 3
	}
	set cyoff -1
	# Format text for display
	set fx [expr {[format "%.3g" [expr {$ux-$xgraphmin}]]+$xgraphmin}]
	set fy [expr {[format "%.3g" [expr {$uy-$ygraphmin}]]+$ygraphmin}]
	# Do display
	set textid [$graph create text \
                [expr {$cx+$cxoff}] [expr {$cy+$cyoff}] \
		-text [format "(%.12g,%.12g)" $fx $fy] \
		-anchor $textanchor -tags position]
	set bbox [$graph bbox $textid]
	eval "$graph create rectangle $bbox -fill white -outline {} \
		-tags position"
	$graph raise $textid
        return [list $fx $fy]
    }
    callback method UnshowPosition {} { graph } {
	$graph delete position
    }
    private method DrawRules { x y } {
	# Note: Imports x and y are in graph (not canvas) coordinates.
	#  If $x=={}, then the vertical rule is drawn on the leftmost
	# edge of the graph.  If $y=={}, then the horizontal rule is
	# drawn on the bottommost edge of the graph.
        set vrulepos $x
        set hrulepos $y
	foreach {cxmin cymin cxmax cymax} \
		[list $lmargin $tmargin \
                  [expr {$lmargin+$plotwidth}] \
                  [expr {$tmargin+$plotheight}]] {}
	set cx {}
	if {![string match {} $x]} {
	    set cx [expr {$x*$xscale+$xoff}]
	}
	set cy {}
	if {![string match {} $y]} {
	    set cy [expr {$y*$yscale+$yoff}]
	}
	if {[string match {} $cx]} { set cx $cxmin }
	if {[string match {} $cy]} { set cy $cymax }
	if {$cx<$cxmin} {set cx $cxmin}
	if {$cx>$cxmax} {set cx $cxmax}
	if {$cy<$cymin} {set cy $cymin}
	if {$cy>$cymax} {set cy $cymax}
	if {$hrulestate} {
	    $graph create line $cxmin $cy $cxmax $cy \
		    -fill $axescolor -width $axeswidth -tags {hrule showpos}
	}
	if {$vrulestate} {
	    $graph create line $cx $cymin $cx $cymax \
		    -fill $axescolor -width $axeswidth -tags {vrule showpos}
	}
    }
    callback method DragInit { tag x y } \
	    { graph but1x but1y lmargin tmargin plotwidth plotheight } {
	set but1x [$graph canvasx $x]
	set but1y [$graph canvasy $y]
	foreach { cx cy } [_PutInBox $but1x $but1y \
		$lmargin $tmargin \
		[expr {$lmargin+$plotwidth}] \
		[expr {$tmargin+$plotheight}]] {}
	switch $tag {
	    hrule {
		$graph raise hrule
		set but1x $cx
		set but1y [lindex [$graph coords hrule] 1]
		## Shift to line center
	    }
	    vrule {
		$graph raise vrule
		set but1x [lindex [$graph coords vrule] 0]
		set but1y $cy
		## Shift to line center
	    }
	}
	$graph raise key  ;# Keep key always on top
    }
    callback method Drag { tag x y } { graph but1x but1y \
	    vrulepos hrulepos lmargin tmargin plotwidth plotheight } {
	set cx [$graph canvasx $x]
	set cy [$graph canvasy $y]
	switch $tag {
	    key   {
                $graph move key [expr {$cx - $but1x}] [expr {$cy - $but1y}]
            }
	    default {
		foreach { cx cy } [_PutInBox $cx $cy \
			$lmargin $tmargin \
			[expr {$lmargin+$plotwidth}] \
			[expr {$tmargin+$plotheight}]] {}
		$graph delete position
		switch $tag {
		    vrule {
			$graph move vrule [expr {$cx - $but1x}]  0
			set vrulepos \
				[lindex [$this ShowPosition $cx $cy 0] 0]
		    }
		    hrule {
			$graph move hrule 0 [expr {$cy - $but1y}]
			set hrulepos \
				[lindex [$this ShowPosition $cx $cy 0] 1]
		    }
		}
	    }
	}
	set but1x $cx
	set but1y $cy
    }
    method DrawAxes {} {
	# Clears canvas, then renders title, axes labels, axes,
	# horizontal and vertical rules, and key.
	#   Also sets {t,b,l,r}margin, plotwidth/plotheight, and
	# {x,y}scale, {x,y}off values.

        # Save key position, *before* updating plotwidth and margins
        set key_pos [$this GetKeyPosition]

        # Check to see if a rule is at or off the edge of the graph,
	# and is so  store {} as the rule position; this will cause
        # '$this DrawRules' to draw the rule at the left/bottom graph
        # edge.
        if {$xscale!=0.} {set slack [expr {1./abs($xscale)}]} \
                else     {set slack 0}
        if {![string match {} $vrulepos] && \
                $vrulepos<=$xgraphmin+$slack || \
                $vrulepos>=$xgraphmax-$slack} {
            set vrulepos {}
        }
        if {$yscale!=0.} {set slack [expr {1./abs($yscale)}]} \
                else     {set slack 0}
        if {![string match {} $hrulepos] && \
                $hrulepos<=$ygraphmin+$slack || \
                $hrulepos>=$ygraphmax-$slack} {
            set hrulepos {}
        }

	# Determine title height
	if {![string match {} $title]} {
	    $graph delete title_test ;# Safety
	    $graph create text 0 0 \
		    -text $title -font $title_font -tags title_test
	    set bbox [$graph bbox title_test]
	    set title_height [expr {1.5*([lindex $bbox 3]-[lindex $bbox 1])}]
	    $graph delete title_test
	} else {
	    set title_height 0
	}

	# Calculate lower bound for margin size from axes coordinate text
	$graph delete margin_test   ;# Safety
	$graph create text 0 0 -text $xgraphmin -font $label_font \
                -anchor sw -tags margin_test
	$graph create text 0 0 -text $xgraphmax -font $label_font \
                -anchor sw -tags margin_test
	set bbox [$graph bbox margin_test]
	set xtext_height [expr {([lindex $bbox 3]-[lindex $bbox 1])*1.2+1}]
	set xtext_halfwidth [expr {([lindex $bbox 2]-[lindex $bbox 0])*0.6+1}]
	## Include a little bit extra
	$graph delete margin_test
	$graph create text 0 0 -text $ygraphmin -font $label_font \
                -anchor sw -tags margin_test
	$graph create text 0 0 -text $ygraphmax -font $label_font \
                -anchor sw -tags margin_test
	set bbox [$graph bbox margin_test]
	set ytext_halfheight [expr {([lindex $bbox 3]-[lindex $bbox 1])*0.6+1}]
	set ytext_width [expr {([lindex $bbox 2]-[lindex $bbox 0])*1.2+1}]
	## Include a little bit extra
	$graph delete margin_test
	## Increase left and top text margin if necessary
	if {$xtext_halfwidth>$ytext_width} {
	    set ytext_width $xtext_halfwidth
	}
	if {$ytext_halfheight>$xtext_height} {
	    set xtext_height $ytext_halfheight
	}

	# Set origin and scaling factors
	# NOTE: Canvas widget has +y downward
	set canwidth [$graph cget -width]
	set canheight [$graph cget -height]
	set lmargin [expr {$canwidth*(1-$graphsize)/2.}]
	set rmargin $lmargin
	set tmargin [expr {$title_height+$canheight*(1-$graphsize)/2.}]
	set bmargin [expr {$tmargin-$title_height}]
	if { $lmargin < $ytext_width }  {
	    set lmargin $ytext_width
	}
	if { $rmargin < $xtext_halfwidth } {
	    set rmargin $xtext_halfwidth
	}
	if { $bmargin < $xtext_height }  {
	    set bmargin $xtext_height
	}
	if { $tmargin < $ytext_halfheight } {
	    set tmargin $ytext_halfheight
	}
	set plotwidth [expr {$canwidth-$lmargin-$rmargin}]
	set plotheight [expr {$canheight-$tmargin-$bmargin}]

	# Set instance variables used to transform from graph coords
	# to display coords:
	set xrange [expr {$xgraphmax-$xgraphmin}]   ;# Temporary
	if {$xrange==0} {set xrange  1}
	set xscale [expr {$plotwidth/$xrange}]
	if {$xscale==0} {set xscale 1}
	set xoff [expr {$lmargin - $xgraphmin * $xscale}]

	set yrange [expr {$ygraphmax-$ygraphmin}]   ;# Temporary
	if {$yrange==0} {set yrange  1}
	set yscale [expr {-$plotheight/$yrange}]
	if {$yscale==0} {set yscale -1}
	set yoff [expr {$tmargin+$plotheight - $ygraphmin * $yscale}]

	set xdispmin $lmargin
	set ydispmin $tmargin
	set xdispmax [expr {$canwidth-$rmargin}]
	set ydispmax [expr {$canheight-$bmargin}]

	# Draw new graph
	$graph delete all
	$this InvalidateDisplay $idlist

	$graph create rectangle 0 0 [expr {2*$canwidth}] $ydispmin \
		-outline {} -fill white -tags axes
	$graph create rectangle 0 $ydispmin $xdispmin $ydispmax \
		-outline {} -fill white -tags axes
	$graph create rectangle $xdispmax $ydispmin \
		[expr {2*$canwidth}] $ydispmax \
		-outline {} -fill white -tags axes
	$graph create rectangle 0 $ydispmax [expr {2*$canwidth}] \
		[expr {2*$canheight}] \
		-outline {} -fill white -tags axes
	## Border matting used to hide curve outliers.  We double outside
	## dimension to hide outliers better at boundary and when
	## resizing the display viewport.

	$graph create rectangle $xdispmin $ydispmin $xdispmax $ydispmax \
		-outline $axescolor -width $axeswidth -tags axes
	$graph create text $xdispmin [expr {$ydispmax+$bmargin*0.1}] \
                -font $label_font \
		-text $xgraphmin -anchor n -tags axes
	$graph create text $xdispmax [expr {$ydispmax+$bmargin*0.1}] \
                -font $label_font \
		-text $xgraphmax -anchor n -tags axes
	$graph create text [expr {$xdispmin-$lmargin*0.1}] $ydispmax \
                -font $label_font \
		-text $ygraphmin -anchor se -tags axes
	$graph create text [expr {$xdispmin-$lmargin*0.1}] $ydispmin \
                -font $label_font \
		-text $ygraphmax -anchor ne -tags axes

	# Render title
	if {![string match {} $title]} {
	    $graph create text \
		    [expr {$canwidth/2.}] \
		    [expr {$tmargin/2.}] \
		    -text $title -font $title_font \
                    -justify center -tags axes
	}

	# Place labels
	$graph create text \
		[expr {($xdispmin+$xdispmax)/2.}] \
		[expr {$canheight-$bmargin*0.1}]  \
                -font $label_font -justify center \
		-text $xlabel -anchor s -tags axes
	$graph create text \
		[expr {$lmargin*0.2}]             \
		[expr {($ydispmin+$ydispmax)/2.}] \
                -font $label_font -justify center \
		-text $ylabel -anchor w -tags axes
	
	# X-axes tics
	set xquart [expr {($xdispmax-$xdispmin)/4.}]
	set x1 [expr {$xdispmin+$xquart}]
	set y1 $ydispmax  ;  set y2 [expr {$y1-1.4*$axeswidth}]
	set y3 $ydispmin  ;  set y4 [expr {$y3+1.4*$axeswidth}]
	$graph create line $x1 $y1 $x1 $y2 \
		-fill $axescolor -width $axeswidth -tags {axes showpos}
	$graph create line $x1 $y3 $x1 $y4 \
		-fill $axescolor -width $axeswidth -tags {axes showpos}
	set x1 [expr {$xdispmax-$xquart}]
	$graph create line $x1 $y1 $x1 $y2 \
		-fill $axescolor -width $axeswidth -tags {axes showpos}
	$graph create line $x1 $y3 $x1 $y4 \
		-fill $axescolor -width $axeswidth -tags {axes showpos}
	set x1 [expr {$xdispmin+2*$xquart}]
	set y1 $ydispmax  ;  set y2 [expr {$y1-2*$axeswidth}]
	set y3 $ydispmin  ;  set y4 [expr {$y3+2*$axeswidth}]
	$graph create line $x1 $y1 $x1 $y2 \
		-fill $axescolor -width $axeswidth -tags {axes showpos}
	$graph create line $x1 $y3 $x1 $y4 \
		-fill $axescolor -width $axeswidth -tags {axes showpos}
	# Y-axes tics
	set yquart [expr {($ydispmax-$ydispmin)/4.}]
	set y1 [expr {$ydispmax-$yquart}]
	set x1 $xdispmin  ;  set x2 [expr {$x1+1.4*$axeswidth}]
	set x3 $xdispmax  ;  set x4 [expr {$x3-1.4*$axeswidth}]
	$graph create line $x1 $y1 $x2 $y1 \
		-fill $axescolor -width $axeswidth -tags {axes showpos}
	$graph create line $x3 $y1 $x4 $y1 \
		-fill $axescolor -width $axeswidth -tags {axes showpos}
	set y1 [expr {$ydispmin+$yquart}]
	$graph create line $x1 $y1 $x2 $y1 \
		-fill $axescolor -width $axeswidth -tags {axes showpos}
	$graph create line $x3 $y1 $x4 $y1 \
		-fill $axescolor -width $axeswidth -tags {axes showpos}
	set y1 [expr {$ydispmax-2*$yquart}]
	set x1 $xdispmin  ;  set x2 [expr {$x1+2*$axeswidth}]
	set x3 $xdispmax  ;  set x4 [expr {$x3-2*$axeswidth}]
	$graph create line $x1 $y1 $x2 $y1 \
		-fill $axescolor -width $axeswidth -tags {axes showpos}
	$graph create line $x3 $y1 $x4 $y1 \
		-fill $axescolor -width $axeswidth -tags {axes showpos}

	# Draw rules
	$this DrawRules $vrulepos $hrulepos

	# Draw key
	$this DrawKey $key_pos
    }

    method NewCurve { curvename } { curvecount id idlist curve color } {
        if {[info exists id($curvename)]} {
	    error "Curve $curvename already exists"
	}
	set curveid $curvecount
	incr curvecount
	lappend idlist $curveid
	set curvecolor $color([expr {$curveid % $color(count)}])
	set id($curvename) $curveid
	array set curve [list                \
		$curveid,name    $curvename  \
		$curveid,color   $curvecolor \
		$curveid,ptsbreak    {}      \
		$curveid,pts         {}      \
		$curveid,lastpt   {-2 0 0}   \
		$curveid,ptsdisp     {}      \
		$curveid,lastdpt     {}      \
                $curveid,dnextpt     0       \
                $curveid,dpiececount 0       \
		$curveid,dispcurrent 1       \
		$curveid,segcount    0       \
		$curveid,hidden      0       \
		$curveid,keyid       {}      \
		$curveid,xmin        {}      \
		$curveid,ymin        {}      \
		$curveid,xmax        {}      \
		$curveid,ymax        {}       ]

	# Update key
	$this DrawKey

	return $curveid
    }
    private method GetCurveId { curvename } { id } {
	# Returns empty string if curvename is invalid
	if {[catch {set id($curvename)} curveid]} {
	    return {}
	}
	return $curveid
    }
    method DeleteCurve { curvename } {
	set curveid [$this GetCurveId $curvename]
	if {[string match {} $curveid]} { return } ;# No such curve

	set index [lsearch -exact $idlist $curveid]
	if {$index >=0} {
	    set idlist [lreplace $idlist $index $index]
	}

	$graph delete tag$curveid
	foreach i [array names curve $curveid,*] {
	    unset curve($i)
	}
	unset id($curvename)

	$this DrawKey      ;# Update key
    }
    method DeleteAllCurves {} {
	# Deletes all data and curves
	foreach name [array names id] {
	    $this DeleteCurve $name
	}
    }
    private method CutCurve { curveid } { curve ptlimit ptlimitcut } {
        if {$ptlimit<1} { return }
        foreach { lastindex x y } $curve($curveid,lastpt) {}
        set newlength [expr {2 * int(floor($ptlimit * $ptlimitcut))}]
	set offset [expr {$lastindex+2-$newlength}]
        if {$offset<=0} { return }
        set curve($curveid,pts) [lrange $curve($curveid,pts) \
                $offset end]
	set brklst {}
	foreach pt $curve($curveid,ptsbreak) {
	    if {$pt>=$offset} {
		lappend brklst [expr $pt-$offset]
	    }
	}
	set curve($curveid,ptsbreak) $brklst
        set xmin $x ; set xmax $x ; set ymin $y ; set ymax $y
        foreach { x y } $curve($curveid,pts) {
            if {$x<$xmin} { set xmin $x }
            if {$x>$xmax} { set xmax $x }
            if {$y<$ymin} { set ymin $y }
            if {$y>$ymax} { set ymax $y }
        }
        set curve($curveid,xmin) $xmin
        set curve($curveid,xmax) $xmax
        set curve($curveid,ymin) $ymin
        set curve($curveid,ymax) $ymax
        set curve($curveid,lastpt) [list [expr {$newlength - 2}] $x $y]
        set curve($curveid,ptsdisp) {}
        set curve($curveid,lastdpt) {}
        set curve($curveid,dnextpt) 0
        set curve($curveid,dpiececount) 0
        set curve($curveid,dispcurrent) 0
    }
    method AddDataPoints { curvename newpts } { id curve ptlimit } {
        # Returns xmin ymin xmax ymax of input data.  (This may
        # be different from extremes of appended data, if ptlimit
        # is smaller than the number of appended points.)
	# Also sets curve($curveid,dispcurrent) to 0 (false).
	# Returns {} if no data input.
	set argcount [llength $newpts]
	if {$argcount<2} { return {} }
	if {$argcount % 2 != 0} {
	    error "List argument must have even number of elements\n\n\
		    usage:\n\t\$instance AddDataPoints curvename list\n\n\
		    Invalid call:\n\t$this AddDataPoints $curvename $newpts"
	}

	if {[catch {set id($curvename)} curveid]} {
	    set curveid [$this NewCurve $curvename]
	}

        # Initialize {x|y}min to "z", {x|y}max to {}, so
        # $x<$xmin and $x>$xmax for *all* numeric $x.
        set xmin [set ymin "z"]
        set xmax [set ymax {}]
	foreach { lastindex x y } $curve($curveid,lastpt) {}
	foreach { tx ty } $newpts {
            if {[string match {} $tx]} {
                # Curve break request; chuck breaks at front and
		# duplicates
		if {$lastindex>=0 && \
			[lindex $curve($curveid,ptsbreak) end] \
			< $lastindex} {
		    # NOTE: If ptsbreak is {}, then the second
		    #  test above will always be true (as a
		    #  string comparison).
		    lappend curve($curveid,ptsbreak) $lastindex
		}
            } else {
                if {[catch {expr {$tx}} x]} {set x 0} ;# Underflow
                if {[catch {expr {$ty}} y]} {set y 0} ;#  protection
                if {$x<$xmin} {set xmin $x}
                if {$x>$xmax} {set xmax $x}
                if {$y<$ymin} {set ymin $y}
                if {$y>$ymax} {set ymax $y}
	        lappend curve($curveid,pts) $x $y
	        incr lastindex 2
            }
        }
        set curve($curveid,lastpt) [list $lastindex $x $y]
        if {[string match {} $xmax]} {
            # No non-empty points entered
            set xmin [set ymin {}]
        } elseif {[string match {} $curve($curveid,xmin)]} {
            set curve($curveid,xmin) $xmin
            set curve($curveid,xmax) $xmax
            set curve($curveid,ymin) $ymin
            set curve($curveid,ymax) $ymax
        } else {
            if {$xmin<$curve($curveid,xmin)} {
                set curve($curveid,xmin) $xmin
            }
            if {$xmax>$curve($curveid,xmax)} {
                set curve($curveid,xmax) $xmax
            }
            if {$ymin<$curve($curveid,ymin)} {
                set curve($curveid,ymin) $ymin
            }
            if {$ymax>$curve($curveid,ymax)} {
                set curve($curveid,ymax) $ymax
            }
        }
	set curve($curveid,dispcurrent) 0
        if {$ptlimit>0 && $lastindex>[expr {2*$ptlimit}]} {
            $this CutCurve $curveid
        }
        if {[string match {} $xmin]} { return }
	return [list $xmin $ymin $xmax $ymax]
    }
    method DeleteAllData {} {
	# Deletes all data, but leaves curve structures
	foreach curveid $idlist {
	    array set curve [list                \
		    $curveid,ptsbreak   {}       \
		    $curveid,pts        {}       \
                    $curveid,lastpt   {-2 0 0}   \
		    $curveid,ptsdisp  {}         \
		    $curveid,lastdpt  {}         \
                    $curveid,dnextpt   0         \
                    $curveid,dpiececount 0       \
		    $curveid,dispcurrent 1       \
		    $curveid,xmin     {}         \
		    $curveid,ymin     {}         \
		    $curveid,xmax     {}         \
		    $curveid,ymax     {}          ]
	    $graph delete tag$curveid
	}
    }
    private method AddDisplayPoints { cid datapts } { \
	    curve xscale yscale xoff yoff \
            xgraphmin ygraphmin xgraphmax ygraphmax } {
	# Converts datapts to display coordinates, and appends to
        # end of last display segment.
	#   To start a new piece, calling routines should lappend an empty
	# list to $curve($cid,ptsdisp) before calling this routine. On
	# exit, this routine will automatically remove the last list elt
	# from $curve($cid,ptsdisp) if it holds less than 2 points (4
	# values).
        #   The return value is a value suitable for passing to PlotCurve
	# for drawing only new points.  Special values: -2 => no
	# new points were added, -3 => last list elt of $curve($cid,ptsdisp)
	# removed.

        set displist [lindex $curve($cid,ptsdisp) end]
        set startindex [llength $displist]

        if {[llength $curve($cid,lastdpt)]==5} {
	    foreach { xprev yprev xprevstat yprevstat previnc } \
		    $curve($cid,lastdpt) {}
	} else {
            # Fresh piece
            set xprev [lindex $datapts 0]
            set yprev [lindex $datapts 1]
	    if {[string match {} $yprev]} {
		# Hmm, no data... Handle case where displist is
		# too short, and return.  This should be a rare
		# circumstance.
		_Assert 0 "ADP -- No data appended to new piece" ;###
		set dlength [llength $displist]
		if {$dlength==2} {
		    # This *really* shouldn't happen!  If lastdpt
		    # is empty, then displist should be too.
		    _Assert 0 "ADP -- Data structure corruption detected" ;###
		    set x [lindex $displist 0]
		    set y [lindex $displist 1]
		    foreach {xstat ystat} [_IsInGraph $x $y \
			    $xgraphmin $ygraphmin \
			    $xgraphmax $ygraphmax] {}
		    set curve($cid,lastdpt) \
			    [list $x $y $xstat $ystat 0]
		}
		if {$dlength<4} {
		    set curve($cid,ptsdisp) \
			    [lreplace $curve($cid,ptsdisp) end end]
		    return -3
		}
		return -2
	    } else {
		foreach {xprevstat yprevstat} [_IsInGraph $xprev $yprev \
			$xgraphmin $ygraphmin $xgraphmax $ygraphmax] {}
		set previnc 0
		set datapts [lreplace $datapts 0 1]
	    }
        }

        if {$previnc} {incr startindex -2}

	_Assert {$startindex>=0} {ADP Bad startindex: $startindex}  ;###

        # Add all points that will display at distinct screen locations
	set xdiff 1.  ; set ydiff 1.   ;# Safety init
	if {$xscale!=0} {set xdiff [expr {1./abs($xscale)}] }
	if {$yscale!=0} {set ydiff [expr {1./abs($yscale)}] }
        foreach { x y } $datapts {
	    if { abs($x-$xprev)<$xdiff && abs($y-$yprev)<$ydiff } {
		continue
	    }
            foreach {xstat ystat} [_IsInGraph $x $y \
                $xgraphmin $ygraphmin $xgraphmax $ygraphmax] {}
            if {$xstat*$xprevstat!=1 && $ystat*$yprevstat!=1} {
                # Some portion of segment inside display window
                if {!$previnc} {
                    # Add previous point to display list
                    set xdisp [expr {$xprev*$xscale+$xoff}]
                    set ydisp [expr {$yprev*$yscale+$yoff}]
                    lappend displist $xdisp $ydisp
                }
                set xdisp [expr {$x*$xscale+$xoff}]
                set ydisp [expr {$y*$yscale+$yoff}]
                lappend displist $xdisp $ydisp
                set previnc 1
            } else {
                set previnc 0
            }
	    set xprev $x         ; set yprev $y
            set xprevstat $xstat ; set yprevstat $ystat
	}

	# Update display point list.  If new list has only 1 point,
	# then strip it out, and leave it in $xprev $yprev but
	# marked as not included.
	set newlength [llength $displist]
	if {$newlength<4} {
	    # New list too short. Strip it from curve($cid,ptsdisp),
	    # but mark curve($cid,lastdpt) appropriately.
	    set previnc 0
	    set startindex -3
	    set curve($cid,ptsdisp) \
		    [lreplace $curve($cid,ptsdisp) end end]
	} else {
	    set curve($cid,ptsdisp) \
		    [lreplace $curve($cid,ptsdisp)   \
		    end end $displist]
	    if {$startindex+4>$newlength} {
		set startindex -2 ;# No new data do display
	    }
	}
        set curve($cid,lastdpt) [list $xprev $yprev \
                $xprevstat $yprevstat $previnc]

	return $startindex
    }
    private method UpdateDisplayList { cid } { curve } {
        # Makes the curve display list up-to-date with respect to
        # the data list.  The return value is the index into the
        # display list necessary to update the graphed display, i.e.,
        # the 'startindex' value appropriate for calling PlotCurve.
        # A value of -1 indicates that the curve should be completely
        # redrawn, and -2 occurs if no points were added to the display
        # list.
        if {$curve($cid,dispcurrent)} { return -2 }

	set startpt $curve($cid,dnextpt)
	set lastpt [lindex $curve($cid,lastpt) 0]
	if {$startpt>$lastpt} {
	    set curve($cid,dispcurrent) 1
	    return -2
	}

	set piececount $curve($cid,dpiececount)
	set breakcount [llength $curve($cid,ptsbreak)]
        if {$piececount>$breakcount} {
	    # Append new points to current piece
	    set startindex [$this AddDisplayPoints $cid \
		    [lrange $curve($cid,pts) $startpt end]]
	    set curve($cid,dnextpt) [expr $lastpt + 2]
	    set curve($cid,dispcurrent) 1
	    return $startindex
	}

	# Otherwise we have to break pts into pieces
	set writecount 0
	set startindex -2 ;# Safety
	set brklist [lrange $curve($cid,ptsbreak) \
		[expr $piececount-1] end]
	if {[llength $brklist]>0} {
	    set bpt [lindex $brklist 0]
	    if {$startpt<=$bpt} {
		if {$piececount<1} {
		    set piececount 1
		    set curve($cid,ptsdisp) [list {}]
		}
		set startindex [$this AddDisplayPoints $cid \
			[lrange $curve($cid,pts) $startpt [expr $bpt+1]]]
		incr writecount
		set startpt [expr $bpt+2]
		set curve($cid,lastdpt) {} ;# End of piece reached
	    } else {
		# Check lastdpt validity
		if {$startpt < $bpt+4} {
		    # $bpt+2 hasn't been processed yet, so start fresh
		    set curve($cid,lastdpt) {} ;# End of piece reached
		}
	    }
	}
	foreach bpt [lrange $brklist 1 end] {
	    incr piececount
	    lappend curve($cid,ptsdisp) {}
	    set startindex [$this AddDisplayPoints $cid \
		    [lrange $curve($cid,pts) $startpt [expr $bpt+1]]]
	    incr writecount
	    set startpt [expr $bpt+2]
	    set curve($cid,lastdpt) {} ;# End of piece reached
	}
	if {$startpt<=$lastpt} {
	    incr piececount
	    lappend curve($cid,ptsdisp) {}
	    set startindex [$this AddDisplayPoints $cid \
		    [lrange $curve($cid,pts) $startpt end]]
	    if {$startindex==-3} {
		# New piece too short to keep
		incr piececount -1
		set startindex -2  ;# No need to redraw last piece
	    } else {
		incr writecount
	    }
	}
	if {$writecount>1} {
	    set startindex -1 ;# Redraw all pieces
	}
	set curve($cid,dpiececount) $piececount
        set curve($cid,dnextpt) [expr $lastpt + 2]
        set curve($cid,dispcurrent) 1
        return $startindex
    }
    private method PlotCurve { curveid update_only} \
	    { graph curve curve_width segmax smoothcurves } {
	# NOTE 1: If update_only is 0, then tag$curveid is
	#  first deleted, and then the entire curve is
	#  redrawn.  Otherwise only that portion of the
	#  curve not yet rendered is drawn.
	if {![info exists curve($curveid,ptsdisp)] || \
                $curve($curveid,hidden)} {
	    $graph delete tag$curveid
	    $this InvalidateDisplay $curveid
	    return
	}
	set startindex [$this UpdateDisplayList $curveid]
	if {!$update_only || $curve($curveid,segcount)>=$segmax} {
	    $graph delete tag$curveid
	    set startindex -1
	}

        if {$startindex == -2} {
            return  ;# Display already up-to-date
        } elseif {$startindex>0} {
            # Draw last section of last piece only
	    incr curve($curveid,segcount)
	    set ptlist [lindex $curve($curveid,ptsdisp) end]
            set ptlist [lrange $ptlist $startindex end]
	    _Assert {[llength $ptlist]>=4 && [llength $ptlist]%2==0} \
		    {PC-a Bad ptlist: $ptlist}       	           ;###
            set gid [eval $graph create line $ptlist \
                    -fill $curve($curveid,color) \
                    -width $curve_width \
                    -smooth $smoothcurves \
                    -tags {[list tag$curveid showpos]}]
            $graph lower $gid
        } else {
            # Redraw all pieces
	    set curve($curveid,segcount) 1
            $graph delete tag$curveid
            foreach ptlist $curve($curveid,ptsdisp) {
		_Assert {[llength $ptlist]>=4 && [llength $ptlist]%2==0} \
			{PC-b Bad ptlist: $ptlist}                     ;###
                set gid [eval $graph create line $ptlist \
                        -fill $curve($curveid,color) \
                        -width $curve_width \
                        -smooth $smoothcurves \
                        -tags {[list tag$curveid showpos]}]
                $graph lower $gid
            }
	}
    }
    method GetCurveList {} {
	# Returns an ordered list of all non-hidden curves (by name)
	set curvelist {}
	foreach curveid $idlist {
	    if {! $curve($curveid,hidden) } {
		lappend curvelist $curve($curveid,name)
	    }
	}
	return $curvelist
    }
    method DrawCurves {{update_only 1}} {
	# Note: We want to draw curves in order
	set curves_drawn 0
	foreach curveid $idlist {
	    if {! $curve($curveid,hidden) } {
		# Curve exists and is not hidden
		$this PlotCurve $curveid $update_only
		incr curves_drawn
	    }
	}
	return $curves_drawn
    }
    private method InvalidateDisplay { curvelist } { curve } {
	# Invalidate display portion of all curves in curvelist.
	# Note: Pass in instance variable $idlist to invalidate
	#  all curves.
	foreach cid $curvelist {
	    array set curve [list \
		    $cid,ptsdisp    {} \
		    $cid,lastdpt    {} \
                    $cid,dnextpt     0 \
                    $cid,dpiececount 0 \
		    $cid,dispcurrent 0 ]
	    if {[lindex $curve($cid,lastpt) 0]<0} {
		set curve($cid,dispcurrent) 1
		## No data points, so ptsdisp is actually up-to-date!
	    }
	}
    }
    private method GetDataExtents {} {
        foreach cid $idlist {
            if {[string match {} $curve($cid,xmin)] || \
                    $curve($cid,hidden) } {
                continue     ;# Empty or hidden curve
            }
            if {![info exists xmin]} {
                # First non-empty curve
		foreach { xmin ymin xmax ymax } [list       \
			$curve($cid,xmin) $curve($cid,ymin) \
			$curve($cid,xmax) $curve($cid,ymax)] {}
            } else {
                # Normal case
		set test $curve($cid,xmin)
		if {$test<$xmin} {set xmin $test}
		set test $curve($cid,ymin)
		if {$test<$ymin} {set ymin $test}
		set test $curve($cid,xmax)
		if {$test>$xmax} {set xmax $test}
		set test $curve($cid,ymax)
		if {$test>$ymax} {set ymax $test}
            }
        }
        if {![info exists xmin]} { 
            # No data, so fill in dummy values
	    foreach { xmin ymin xmax ymax } { 0 0 1 1 } {}
        }
        if {$xmax<=$xmin} { set xmax \
		[expr {$xmin+(1.000001)*(abs($xmin)+1.e-6)}]}
        if {$ymax<=$ymin} { set ymax \
		[expr {$ymin+(1.000001)*(abs($ymin)+1.e-6)}]}

        return [list $xmin $ymin $xmax $ymax]
    }
    private method RoundExtents {xmin ymin xmax ymax} {
	# Adjust to "round" numbers
	set xdelta [expr {($xmax-$xmin)*0.999}]  ;# Allow a *little* slack
	if {$xdelta<=0.0} {set xdelta 1. } ;# Safety 
	set xrangescale [expr {pow(10,ceil(-log10($xdelta)))}]
	set xmin [expr {(floor($xmin*$xrangescale))/$xrangescale}]
	set xmax [expr {(ceil($xmax*$xrangescale))/$xrangescale}]
	set ydelta [expr {($ymax-$ymin)*0.999}]  ;# Allow a *little* slack
	if {$ydelta<=0.0} {set ydelta 1. } ;# Safety 
	set yrangescale [expr {pow(10,ceil(-log10($ydelta)))}]
	set ymin [expr {(floor($ymin*$yrangescale))/$yrangescale}]
	set ymax [expr {(ceil($ymax*$yrangescale))/$yrangescale}]
        
	# Round and save for future use
	set xtestmin [format "%.12g" $xmin]
	set xtestmax [format "%.12g" $xmax]
	if { $xtestmax<=$xtestmin } {
	    # Too much rounding.  Try again
	    set xtestmin [format "%.15g" $xmin]
	    set xtestmax [format "%.15g" $xmax]
	}
	set xmin $xtestmin
	set xmax $xtestmax
	set ytestmin [format "%.12g" $ymin]
	set ytestmax [format "%.12g" $ymax]
	if { $ytestmax<=$ytestmin } {
	    # Too much rounding.  Try again
	    set ytestmin [format "%.15g" $ymin]
	    set ytestmax [format "%.15g" $ymax]
	}
	set ymin $ytestmin
	set ymax $ytestmax

	# Safety
        global tcl_precision
        set eps [expr {pow(10.,-$tcl_precision)}]
	if {$xmax<=$xmin} {
	    set xmax [expr {$xmin+$eps*abs($xmin)}]
	}
	if {$ymax<=$ymin} {
	    set ymax [expr {$ymin+$eps*abs($ymin)}]
	}

	return [list $xmin $ymin $xmax $ymax]
    }
    method SetGraphLimits {{xmin {}} {ymin {}} {xmax {}} {ymax {}}} {
        # Sets the graph numerical limits.  Any blank values are
        # filled in using data values.  This routine does not update
	# the display.
        #  If any change is made, all curve displays are marked invalid,
        # and 1 is returned.  If no change is made, then 0 is returned.
        foreach {oxmin oymin oxmax oymax} \
                [list $xgraphmin $ygraphmin $xgraphmax $ygraphmax] {}

        if {[string match {} $xmin] || [string match {} $ymin] \
                || [string match {} $xmax] || [string match {} $ymax]} {
            foreach {xgraphmin ygraphmin xgraphmax ygraphmax} \
                    [$this GetDataExtents] {}
	    foreach { xgraphmin ygraphmin xgraphmax ygraphmax } \
		    [$this RoundExtents \
		     $xgraphmin $ygraphmin $xgraphmax $ygraphmax] {}
	    if {![string match {} $xmin]} { set xgraphmin $xmin }
	    if {![string match {} $ymin]} { set ygraphmin $ymin }
	    if {![string match {} $xmax]} { set xgraphmax $xmax }
	    if {![string match {} $ymax]} { set ygraphmax $ymax }
	} else {
	    if {$xmin<=$xmax} {
		set xgraphmin $xmin ; set xgraphmax $xmax
	    } else {
		set xgraphmin $xmax ; set xgraphmax $xmin
	    }
	    if {$ymin<=$ymax} {
		set ygraphmin $ymin ; set ygraphmax $ymax
	    } else {
		set ygraphmin $ymax ; set ygraphmax $ymin
	    }
	}

        # Safety adjustment
        global tcl_precision
        set eps [expr {pow(10.,-$tcl_precision)}]
        if {$xgraphmax<=$xgraphmin} {
            set xgraphmax [expr {$xgraphmin+$eps*abs($xgraphmin)}]
        }
        if {$ygraphmax<=$ygraphmin} {
            set ygraphmax [expr {$ygraphmin+$eps*abs($ygraphmin)}]
        }

        if {$oxmin!=$xgraphmin || $oymin!=$ygraphmin || \
                $oxmax!=$xgraphmax || $oymax!=$ygraphmax} {
            # Mark all curve displays invalid
	    $this InvalidateDisplay $idlist
            return 1
        }
        # Otherwise, no change
        return 0
    }
    method GetGraphLimits {} {xgraphmin ygraphmin xgraphmax ygraphmax} {
        return [list $xgraphmin $ygraphmin $xgraphmax $ygraphmax]
    }
    method RefreshDisplay {} {
	$this DrawAxes
	$this DrawCurves 0
    }
    method Reset {} {
	# Re-initializes most of the graph properties to initial defaults
	$this DeleteAllCurves
	$this SetGraphLimits 0 0 1 1
	set hrule {} ; set vrule {}
	set title  {}
	set xlabel {}
	set ylabel {}
	$this DrawAxes
    }
    method PrintData { curvepat } {
	foreach curvename [array names id $curvepat] {
	    set curveid $id($curvename)
	    foreach i [lsort -ascii [array names curve $curveid,*]] {
		puts "$i = $curve($i)"
	    }
	}
	return
    }
    method HasGui {} {
        return 0
    }
    method WinDestroy { w } { winpath } {
	if {[string compare $winpath $w]!=0} {
	    return    ;# Child destroy event
	}
	$this Delete
    }
    Destructor {
	# Notify client of impending destruction
	catch {eval $callback $this DELETE {$delete_callback_arg}}
        Oc_EventHandler Generate $this Delete
	if {[info exists winpath] && [winfo exists $winpath]} {
            # Remove <Destroy> binding
	    bind $winpath <Destroy> {}
	    # Destroy instance frame, all children, and bindings
	    destroy $winpath
	}
    }
}

# Utility proc that determines if the point (x,y) is in the rectangle
# (xmin,xmax)x(ymin,ymax).  This is made a Tcl proc rather than an
# Oc_Class method to cut down on call overhead (this routine is called
# inside the curve scaling routine on each potential display point).
;proc _IsInGraph { x y xmin ymin xmax ymax } {
    if {$x<$xmin}             {set xpos -1} \
            elseif {$x>$xmax} {set xpos  1} \
            else              {set xpos  0}
    if {$y<$ymin}             {set ypos -1} \
            elseif {$y>$ymax} {set ypos  1} \
            else              {set ypos  0}
    return [list $xpos $ypos]
}
# Similar to the preceding proc, this one adjusts $x and $y as necessary
# to fit into the specified bounding box.
;proc _PutInBox { x y xmin ymin xmax ymax } {
    if {$x<$xmin}     { set x $xmin } \
    elseif {$x>$xmax} { set x $xmax }
    if {$y<$ymin}     { set y $ymin } \
    elseif {$y>$ymax} { set y $ymax }
    return [list $x $y]
}

# Debugging proc.  If 'expr $test' is false, then raises
# error with message "$errmsg".
;proc _Assert { test errmsg } {
    if {![uplevel 1 expr $test]} {
	# Kill links to shut off incoming traffic
	foreach link [Net_Link Instances] {
	    catch {$link Delete}
	}
	after 1000   ;# Pause briefly
	uplevel 1 error \"$errmsg\"
    }
}
